(* Copyright 2001, 2002 b8_bavard, b8_fee_carabine, INRIA *)
(*
    This file is part of mldonkey.

    mldonkey is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    mldonkey is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with mldonkey; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

open Int64ops
open Printf2
open Md4
open Options
  
open BasicSocket
open TcpBufferedSocket

open CommonSwarming
open CommonSources
open CommonDownloads
open CommonInteractive
open CommonSearch
open CommonGlobals
open CommonComplexOptions
open CommonFile
open CommonClient
open CommonComplexOptions
open CommonTypes
open CommonOptions

open DonkeyMftp
open DonkeyImport
open DonkeyProtoCom
open DonkeyTypes
open DonkeyOptions
open DonkeyComplexOptions
open DonkeyGlobals
open DonkeyChunks
open DonkeyReliability
open DonkeyStats
    
(* Sort files in clients file queue in order of priority and percentage downloaded
   This way higher priority files will be asked/downloaded first if the client does have more
   than one file to offer.
   Only sort if client_block is not set.
   Once the block has been finished allow changing order.
*)

let sort_file_queue c =
  match c.client_download with
    Some _ -> ()
  | None ->
      match c.client_file_queue with
        [] -> ()
      | [ (file, chunks, up) ] ->
          if !verbose_download || c.client_debug then begin
              lprintf "sort_file_queue: single file. client(%d): %s, file(%d): %s\n" (client_num c) c.client_name (file_num file) (file_best_name file);
            end
      | (file, chunks, up) :: _ ->
          let fn = file_num file in
          if !verbose_download || c.client_debug then begin
              lprintf "sort_file_queue: multiple files. client(%d): %s, file(%d): %s\n" (client_num c) c.client_name (file_num file) (file_best_name file);
            end;
          c.client_file_queue <- List.stable_sort (fun (f1, _, _) (f2, _, _) ->
              let v = file_priority f2 - file_priority f1 in
              if v <> 0 then v else
              let s1 = if (file_size f1) > Int64.zero then
                  Int64.to_int (Int64.div (Int64.mul (file_downloaded f1) (Int64.of_int 100)) (file_size f1))
                else 0 in
              let s2 = if (file_size f2) > Int64.zero then
                  Int64.to_int (Int64.div (Int64.mul (file_downloaded f2) (Int64.of_int 100)) (file_size f2))
                else 0 in
              s2 - s1
          ) c.client_file_queue;
          match c.client_file_queue with
            [] -> ()
          | (file, chunks, _) :: _ ->
              if (file_num file) <> fn then begin
                  if !verbose_download || c.client_debug then begin
                      lprintf "sort_file_queue: queue change. client(%d): %s, file(%d): %s\n" (client_num c) c.client_name (file_num file) (file_best_name file);
                    end;
(*
(*                  c.client_chunks <- chunks; *)
(*                  c.client_all_chunks <- String.make file.file_nchunks '0'; *)
(*                  c.client_zones <- []; *)
                  for i = 0 to file.file_nchunks - 1 do
                    if c.client_chunks.(i)  then
                      c.client_all_chunks.[i] <- '1';
done;
  *)
                end

let remove_client_slot c =
  set_client_has_a_slot (as_client c) false;
  client_send c (
    let module M = DonkeyProtoClient in
    let module Q = M.CloseSlot in
    M.CloseSlotReq Q.t);
  c.client_upload <- None

let unshare_file file =
  match file.file_shared with
    None -> ()
  | Some s -> 
      file.file_shared <- None;
      decr nshared_files;
      CommonShared.shared_calculate_total_bytes ();
      (try Unix32.close  (file_fd file) with _ -> ());
      
      begin
        H.iter (fun c ->
            match c.client_upload with
              Some { up_file = f } when f == file ->
                remove_client_slot c
            | _ -> ()
        ) clients_by_kind
      end

let declare_completed_file file = 
  DonkeyShare.remember_shared_info file (file_disk_name file);
  file_completed (as_file file);
  
  unshare_file file;
  (*ignore (CommonShared.new_shared "completed" 0 (
      file_best_name file )
    (file_disk_name file));*)
  (try
      let format = CommonMultimedia.get_info
          (file_disk_name file) in
      file.file_format <- format
    with _ -> ())
  
(** What to do when a file is finished
  @param file the finished file
*)         
let download_finished file = 
  if List.memq file !current_files then begin      
      current_files := List2.removeq file !current_files;
      declare_completed_file file
    end

(** Check if a file is finished or not.
  A file is finished if all blocks are verified.
  @param file The file to check status
*)
let check_file_downloaded file = 
  match file_state file with
    FileCancelled | FileShared | FileDownloaded -> ()
  | _ ->
      match file.file_swarmer with
        None -> ()
      | Some swarmer ->
          let bitmap = Int64Swarmer.verified_bitmap swarmer in
(*          lprintf "Verified bitmap: [%s]\n" bitmap; *)
          let rec iter i =
            if i =  String.length bitmap then true
            else
            if bitmap.[i] = '3' then iter (i+1) else false
          in
          let verified = iter 0 in
          
          let downloaded = Int64Swarmer.downloaded swarmer in
          if file_downloaded file <> downloaded then begin
              lprintf "ERROR: file_downloaded file (%Ld) <> downloaded swarmer (%Ld)\n"
                (file_downloaded file) downloaded
            end;
          
          if verified then begin
              if (file_size file <> downloaded)
              then
                lprintf "DonkeyOneFile: Downloaded size differs after complete verification\n";
              download_finished file
            end
            
let check_files_downloaded () =
  List.iter check_file_downloaded !current_files;
  Hashtbl.iter (fun file_md4 file ->
      match file.file_shared with 
        Some _ -> ()
      | None ->
          match file.file_swarmer with
            None -> ()
          | Some swarmer ->
              let bitmap = Int64Swarmer.verified_bitmap swarmer in
              let rec iter i len =
                if i < len then
                  if bitmap.[i] = '3' then
                    DonkeyShare.must_share_file file
                  else
                    iter (i+1) len
              in
              iter 0 (String.length bitmap)
  ) files_by_md4

let add_client_chunks c file client_chunks =
  match file.file_swarmer with
    None -> failwith "add_client_chunks: no swarmer"
  | Some swarmer ->
      let rec iter list =
        match list with
          (f, chunks, up) :: tail ->
            if f != file then iter tail
            else begin
                Int64Swarmer.update_uploader up
                (AvailableBoolBitmap client_chunks);
                Array.blit client_chunks 0 chunks 0 (Array.length chunks)
              end
            
        | [] ->
            let up = Int64Swarmer.register_uploader swarmer (as_client c) 
              (AvailableBoolBitmap client_chunks) in
            c.client_file_queue <-  c.client_file_queue @
              [file, client_chunks, up]
      in
      iter c.client_file_queue
      

(* let next_file _ = failwith "next_file not implemented" *)
      
(* clean_client_zones: clean all structures related to downloads when
   a client disconnects *)
let clean_current_download c = 
  match c.client_download with
    None -> ()
  | Some (file, up) ->
      Int64Swarmer.clear_uploader_block up;
      Int64Swarmer.clear_uploader_ranges up;
      c.client_download <- None

let send_get_range_request c file ranges = 
  match c.client_source.DonkeySources.source_sock with
  | Connection sock ->
      
      set_rtimeout sock !queue_timeout;
      let module M = DonkeyProtoClient in
      let module Q = M.QueryBloc in
      let msg, len =
        match ranges with
          [x1,y1,_] ->
            {
              Q.md4 = file.file_md4;
              Q.start_pos1 = x1;
              Q.end_pos1 = y1;
              Q.start_pos2 = zero;
              Q.end_pos2 = zero;
              Q.start_pos3 = zero;
              Q.end_pos3 = zero;
            }, y1 -- x1
        
        | [x1,y1,_; x2,y2,_] ->
            {
              Q.md4 = file.file_md4;
              Q.start_pos1 = x1;
              Q.end_pos1 = y1;
              Q.start_pos2 = x2;
              Q.end_pos2 = y2;
              Q.start_pos3 = zero;
              Q.end_pos3 = zero;
            }, y1 -- x1
        
        | [x1,y1,_; x2,y2,_; x3,y3,_ ] ->
            {
              Q.md4 = file.file_md4;
              Q.start_pos1 = x1;
              Q.end_pos1 = y1;
              Q.start_pos2 = x2;
              Q.end_pos2 = y2;
              Q.start_pos3 = x3;
              Q.end_pos3 = y3;
            }, y1 -- x1
        
        | _ -> assert false
      in
      let msg = M.QueryBlocReq msg in
      set_read_power sock (!!upload_power + maxi 0 (file_priority file));
(*      lprintf "QUEUE DOWNLOAD REQUEST\n"; *)
(*      CommonUploads.queue_download_request (fun _ ->  *)
          client_send c msg 
(*          ) (Int64.to_int len) *)
  | _ -> assert false
      
let rec get_from_client c =
  match c.client_download with
    None ->
(*      lprintf "get_from_client: no download\n"; *)
      begin
        match c.client_file_queue with
          [] -> 
            
(*            lprintf "get_from_client: no more file\n"; *)
            if not (client_has_a_slot (as_client c)) then begin
(*                connection_delay c.client_connection_control; *)
                match c.client_source.DonkeySources.source_sock with
                  Connection sock ->
                    TcpBufferedSocket.close sock
                      (Closed_for_error "No file to download");
                | _ -> ()
              end
        
        | (file, chunks, up) :: tail ->

(* Should we start a download without asking for a slot first ?? *)
(*            lprintf "get_from_client: next file\n"; *)
            c.client_download <- Some (file,up);
            get_from_client c
      end
  
  | Some (file,up) ->
      
      try
        let b = Int64Swarmer.current_block up in
        let ranges = Int64Swarmer.current_ranges up in
        let before_find_range = List.length ranges in

(*        lprintf "WAITING FOR %d BLOCS\n" before_find_range; *)
        if before_find_range < 3 then
          let rec iter n =
            if n < 3 then
              try
                ignore (Int64Swarmer.find_range up);
                iter (n+1)
              with 
                Not_found -> n
            else n
          in
          let after_find_range = iter before_find_range in
          if after_find_range > before_find_range then 
            
            let ranges = Int64Swarmer.current_ranges up in
            send_get_range_request c file ranges;
            
          else
(* No new range to download in this block *)
              match ranges with
                [] ->
                  raise Not_found (* will query the next block *)
                  
              | _ -> 
(* Wait for the already requested ranges before requesting the next block *)
                  ()                  
              
        else
(* We already have 3 ranges in the current block *)
          ()
      with Not_found ->
(*          lprintf "get_from_client: no range\n"; *)
          try
            let swarmer = Int64Swarmer.uploader_swarmer up in       
            (try Int64Swarmer.verify_one_chunk swarmer with _ -> ());
            let b = Int64Swarmer.find_block up in
            get_from_client c
            
          with Not_found ->
(*              lprintf "get_from_client: no block\n"; *)
              match Int64Swarmer.current_ranges up with
                [] ->
(* We have nothing to wait for in the current file *)
                  begin
                    
(*                    lprintf "get_from_client: no expected ranges\n"; *)
                    
                    c.client_download <- None;
                    Int64Swarmer.unregister_uploader up;
                    match c.client_file_queue with
                      [] -> assert false
                    | _ :: tail -> 
(* We can go to next file now *)
                        c.client_file_queue <- tail;
                        get_from_client c
                  end
              | _ -> 
(* We are still waiting for the previous requested ranges from this file *)
                  ()
                  
      
(* start_download: ask for a slot in the queue of the remote client,
  or start querying blocks if already in the queue *)
let request_slot c = 
    if c.client_slot = SlotNotAsked then begin
      if !verbose_download then begin
          lprintf "start_download...\n";
        end;
      do_if_connected c.client_source.DonkeySources.source_sock (fun sock ->
          sort_file_queue c;
          match c.client_file_queue with
            [] -> ()
          | (file, _,_ ) :: _ ->
              client_send c (
                let module M = DonkeyProtoClient in
                let module Q = M.JoinQueue in
                M.JoinQueueReq (Some file.file_md4));
              c.client_slot <- SlotAsked;
      )
    end
  
let block_received c md4 begin_pos bloc bloc_pos bloc_len = 
    
  match c.client_download with
    None -> 
      lprintf "DonkeyOneFile.block_received: block received but no file !\n  Received: %s\n" (Md4.to_string md4)
  | Some (file, up) ->
      
      if file.file_md4 <> md4 then begin
          lprintf "DonkeyOneFile.block_received: block for bad file\n  Received: %s\n  Expected: %s\n" (Md4.to_string md4) (Md4.to_string file.file_md4)
        end else begin
          DonkeySources.set_request_result c.client_source file.file_sources File_upload;
          
          c.client_rating <- c.client_rating + 10;
          
          set_client_state c (Connected_downloading (file_num file));
          let len64 = Int64.of_int bloc_len in
          count_download c file len64;

(* TODO: verify the received data has been requested *)
          
          let swarmer = Int64Swarmer.uploader_swarmer up in
          let old_downloaded = Int64Swarmer.downloaded swarmer in
          
          begin
            try
              Int64Swarmer.received up begin_pos bloc bloc_pos bloc_len
            with
            | e ->
                let m =
(*		      Printf.sprintf "File %s begin_pos=%s bloc_begin=%d bloc_len=%d:\nError %s while writing block%s\n" (file_best_name file) (Int64.to_string begin_pos) t.Q.bloc_begin t.Q.bloc_len (Printexc2.to_string e)  *)
                  (match e with 
                      Unix.Unix_error (Unix.ENOSPC, _, _) -> " (Disk full?)"
                    | _ -> "") in
(*                    Printf2.lprint_string m; *)
                CommonEvent.add_event (Console_message_event m);
                if e <> End_of_file then begin
                    let m = "File paused.\n" in
                    Printf2.lprint_string m;
                    CommonEvent.add_event (Console_message_event m);
                    file_pause (as_file file);
                    raise e
                  end
          
          end;

(*            List.iter Int64Swarmer.alloc_range c.client_ranges; *)
          let new_downloaded = 
            Int64Swarmer.downloaded swarmer in
          c.client_downloaded <- c.client_downloaded ++ (
            new_downloaded -- old_downloaded);
(*
          if not (List.mem c.client_ip bb.block_contributors) then
            bb.block_contributors <- c.client_ip :: 
            bb.block_contributors;
*)          
          
          
          if new_downloaded -- old_downloaded < len64 then
              if !verbose_download then begin
                lprintf "ALREADY RECEIVED: %Ld < %Ld\n"
              (new_downloaded -- old_downloaded) len64;          
              end;
          get_from_client c
        end    
        
let search_found filter search md4 tags =
  let file_name = ref "" in
  let file_size = ref Int64.zero in
  let availability = ref 0 in
  let new_tags = ref [] in
  List.iter (fun tag ->
      match tag with
        { tag_name = "filename"; tag_value = String s } -> file_name := s
      | { tag_name = "size"; tag_value = Uint64 v } -> file_size := v
      | { tag_name = "availability";
	  tag_value = (Uint64 v| Fint64 v) } ->
          availability := Int64.to_int v;  new_tags := tag :: !new_tags
      | _ -> new_tags := tag :: !new_tags
  ) tags;
(*  TODO INDEX try
    let rs = DonkeyIndexer.find_result md4 in
    CommonInteractive.search_add_result filter search rs.result_result; 
(* TODO ADD AVAILABILITY *)
    let doc = rs.result_index in
    let result = Store.get store doc in
    if not (List.mem !file_name result.result_names) then begin
        DonkeyIndexer.add_name result !file_name;
        result.result_names <- !file_name :: result.result_names
      end
  with _ -> *)
      match result_of_file md4 tags with
        None -> ()
      | Some r ->
      try
(* TODO INDEX       let rs = DonkeyIndexer.index_result new_result in 
        let doc = rs.result_index in
        CommonInteractive.search_add_result filter search rs.result_result;
       let result = Store.get store doc in
*)
        CommonInteractive.search_add_result filter search r;
        ()
      with _ ->  (* the file was probably filtered *)
          ()
